import uuid

from rethinkdb import r

from eyja.interfaces.db import BaseStorageClient

from .config import RethinkDBConfig


class RethinkDBClient(BaseStorageClient):
    _connection = None
    _config: RethinkDBConfig
    _config_cls = RethinkDBConfig

    async def init(self):
        r.set_loop_type('asyncio')
        self._connection = await r.connect(
            host=self._config.host,
            port=self._config.port,
            password=self._config.password,
        )

        db_list = await r.db_list().run(self._connection)
        self._buckets.extend(self._config.db)
        for db in self._config.db:
            if db not in db_list:
                await r.db_create(db).run(self._connection)

    async def init_table(self, db, table, indexes):
        table_list = await r.db(db).table_list().run(self._connection)
        if table not in table_list:
            await r.db(db).table_create(table).run(self._connection)

        index_list = await r.db(db).table(table).index_list().run(self._connection)
        for index in indexes:
            if index not in index_list:
                await r.db(db).table(table).index_create(index).run(self._connection)
                await r.db(db).table(table).index_wait(index).run(self._connection)

    async def save(self, obj, object_space, object_type):
        await self.init_table(object_space, object_type, obj._indexes)

        if not obj.object_id:
            obj.object_id = str(uuid.uuid4())

            await r.db(object_space).table(object_type).insert(obj.data).run(self._connection)
            return

        data = obj.data
        data.pop('created_at')
        
        await r.db(
            object_space
        ).table(
            object_type
        ).filter(
            {'object_id': obj.object_id}
        ).limit(1).update(data).run(self._connection)

    async def delete(self, obj, object_space, object_type):
        await self.init_table(object_space, object_type, obj._indexes)

        await r.db(
            object_space
        ).table(
            object_type
        ).filter(
            {'object_id': obj.object_id}
        ).limit(1).delete().run(self._connection)

    async def delete_all(self, obj, object_space, object_type, filter):
        await self.init_table(object_space, object_type, obj._indexes)

        await r.db(
            object_space
        ).table(
            object_type
        ).filter(
            filter
        ).delete().run(self._connection)

    async def get(self, obj_cls, object_space, object_type, object_id):
        await self.init_table(object_space, object_type, obj_cls._indexes)

        result = await r.db(
            object_space
        ).table(
            object_type
        ).get_all(
            object_id,
            index='object_id'
        ).limit(1).run(self._connection)

        items = [item async for item in result]
        if len(items) < 1:
            return None

        return obj_cls(**items[0])

    async def get_from_index(self, obj_cls, object_space, object_type, value, index):
        await self.init_table(object_space, object_type, obj_cls._indexes)

        result = await r.db(
            object_space
        ).table(
            object_type
        ).get_all(
            *value if isinstance(value, list) else [value],
            index=index
        ).run(self._connection)

        items = [item async for item in result]
        if len(items) < 1:
            return None

        return [obj_cls(**e) for e in items]

    async def find(self, obj_cls, object_space, object_type, filter):
        await self.init_table(object_space, object_type, obj_cls._indexes)

        if filter.order_direction == 'asc':
            order_by_expr = r.asc(filter.order_by)
        else:
            order_by_expr = r.desc(filter.order_by)

        result = await r.db(
            object_space
        ).table(
            object_type
        ).order_by(
            index=order_by_expr
        ).filter(
            filter.fields
        ).skip(
            filter.page_size*filter.page_no
        ).limit(
            filter.page_size
        ).run(self._connection)

        items = [item async for item in result]
        return [obj_cls(**item) for item in items]

    async def count(self, obj_cls, object_space, object_type, filter):
        await self.init_table(object_space, object_type, obj_cls._indexes)

        return await r.db(
            object_space
        ).table(
            object_type
        ).filter(
            filter.fields
        ).count().run(self._connection)



from unittest import IsolatedAsyncioTestCase

from rethinkdb import r

from eyja.main import Eyja
from eyja.hubs import DataHub
from eyja.interfaces.db import BaseStorageModel


class RethinkDBClientTest(IsolatedAsyncioTestCase):
    rethinkdb_connection = None

    config = '''
        storages:
            rethinkdb:
                port: 30001
                password: 123456
                db:
                -   test_db
    '''

    class TestModel(BaseStorageModel):
        _namespace = ':::test_table'
        _indexes = [
            'test_str'
        ] + BaseStorageModel._indexes
        
        test_str: str

    async def asyncSetUp(self):
        r.set_loop_type('asyncio')
        self.rethinkdb_connection = await r.connect(host='localhost', port=30001, password='123456')

    async def find_items(self, db, table, **filter):
        result = await r.db(db).table(table).filter(filter).run(self.rethinkdb_connection)
        return [item async for item in result]

    async def test_connection(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        db_list = await r.db_list().run(self.rethinkdb_connection)

        self.assertTrue(Eyja.is_initialized())
        self.assertIn('test_db', db_list)

    async def test_save_model(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        obj = self.TestModel(
            test_str='Test123'
        )
        await obj.save()

        table_list = await r.db('test_db').table_list().run(self.rethinkdb_connection)
        index_list = await r.db('test_db').table('test_table').index_list().run(self.rethinkdb_connection)
        saved_data = await self.find_items('test_db', 'test_table', test_str='Test123')

        self.assertIn('test_table', table_list)
        self.assertIn('test_str', index_list)
        self.assertEqual(len(saved_data), 1)

    async def test_delete_model(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        obj = self.TestModel(
            test_str='Test1'
        )
        await obj.save()
        await obj.delete()

        saved_data = await self.find_items('test_db', 'test_table', test_str='Test1')
        self.assertEqual(len(saved_data), 0)

    async def test_delete_all_models(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        await self.TestModel(test_str='Test2').save()
        await self.TestModel(test_str='Test2').save()
        await self.TestModel(test_str='Test2').save()
        await self.TestModel(test_str='Test22').save()
        await self.TestModel(test_str='Test22').save()

        await self.TestModel.delete_all({'test_str': 'Test2'})

        self.assertEqual(len(await self.find_items('test_db', 'test_table', test_str='Test2')), 0)
        self.assertEqual(len(await self.find_items('test_db', 'test_table', test_str='Test22')), 2)

    async def test_get_model(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        obj = self.TestModel(
            test_str='Test3'
        )
        await obj.save()

        get_obj = await self.TestModel.get(obj.object_id)

        self.assertEqual(get_obj.test_str, obj.test_str)

    async def test_find_models(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        await self.TestModel(test_str='Test4').save()
        await self.TestModel(test_str='Test4').save()
        await self.TestModel(test_str='Test4').save()
        await self.TestModel(test_str='Test44').save()
        await self.TestModel(test_str='Test44').save()

        self.assertEqual(len(await self.TestModel.find({'test_str': 'Test4'})), 3)
        self.assertEqual(len(await self.TestModel.find({'test_str': 'Test44'})), 2)

    async def test_find_models(self):
        await Eyja.reset()
        await Eyja.init(
            config=self.config,
            plugins=['eyja_rethinkdb']
        )

        await self.TestModel(test_str='Test5').save()
        await self.TestModel(test_str='Test5').save()
        await self.TestModel(test_str='Test5').save()
        await self.TestModel(test_str='Test55').save()
        await self.TestModel(test_str='Test55').save()

        self.assertEqual(await self.TestModel.count({'test_str': 'Test5'}), 3)
        self.assertEqual(await self.TestModel.count({'test_str': 'Test55'}), 2)



from setuptools import find_packages, setup


setup(
    name='eyja-rethinkdb',
    zip_safe=True,
    version='0.1.10',
    description='RethinkDB Plugin for Eyja',
    url='https://gitlab.com/public.eyja.dev/eyja-rethinkdb',
    maintainer='Anton Berdnikov',
    maintainer_email='agratoth@yandex.ru',
    packages=find_packages(),
    package_dir={'eyja_rethinkdb': 'eyja_rethinkdb'},
    install_requires=[
        'eyja-internal>==0.4.4',
        'rethinkdb>=2.4.8',
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires='>=3.8',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)
